@@ -13,6 +13,7 @@
 #    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 #    See the License for the specific language governing permissions and
 #    limitations under the License.
+from collections.abc import Iterable
 import math
 import numpy as np
 import numbers
@@ -21,6 +22,8 @@ import pandas as pd
 from neo4j import GraphDatabase
 
 from bluegraph.core.io import GraphProcessor, PandasPGFrame
+from bluegraph.core.utils import normalize_to_set
+from bluegraph.exceptions import (BlueGraphException, BlueGraphWarning)
 
 
 def execute(driver, query):
@@ -67,18 +70,64 @@ def _generate_property_repr(properties, prop_types=None):
     props = []
     for k, v in properties.items():
         if k not in ["@id", "@type"]:
-            quote = "'"
-            if prop_types[k] == "numeric" and not math.isnan(v):
-                quote = ""
-            props.append("{}: {}{}{}".format(
-                k, quote, preprocess_value(v), quote))
+            if isinstance(v, str):
+                # create a string property
+                quote = "'"
+                props.append("{}: {}{}{}".format(
+                    k.replace(".", "_"), quote,
+                    str(preprocess_value(v)).replace("'", "\\'"), quote))
+            elif isinstance(v, Iterable):
+                # create a list property
+                for vv in v:
+                    values = []
+                    if isinstance(vv, float) and math.isnan(vv):
+                        pass
+                    else:
+                        values.append("'{}'".format(preprocess_value(vv)))
+                if len(values) > 0:
+                    props.append("{}: [{}]".format(
+                        k.replace(".", "_"), ", ".join(values)))
+            elif prop_types[k] == "numeric" and not math.isnan(v):
+                # create a numerical property
+                props.append("{}: {}".format(
+                    k.replace(".", "_"), preprocess_value(v)))
     return props
 
 
+def labels_from_types(properties):
+    if "@type" in properties:
+        type_set = normalize_to_set(properties["@type"])
+        return list(type_set)
+
+
 def pgframe_to_neo4j(pgframe=None, uri=None, username=None, password=None,
-                     driver=None, node_label=None, edge_label=None,
-                     directed=True, batch_size=10000):
+                     driver=None, node_label=None, edge_label=None, directed=True,
+                     node_types_as_labels=False,
+                     edge_types_as_labels=False,
+                     batch_size=10000):
     """Write the property graph to the Neo4j databse."""
+    if node_label is None:
+        if node_types_as_labels is False or not pgframe.has_node_types():
+            raise BlueGraphException(
+                "Cannot create a Neo4j graph without node labels: "
+                "node label is not provided "
+                "and 'node_types_as_labels' is either set to False "
+                "or the nodes do not have types")
+
+    if edge_label is None:
+        if edge_types_as_labels is False or not pgframe.has_edge_types():
+            raise BlueGraphException(
+                "Cannot create a Neo4j graph without edge labels: "
+                "edge label is not provided "
+                "and 'edge_types_as_labels' is either set to False "
+                "or the edges do not have types")
+    else:
+        if edge_types_as_labels is True and pgframe.has_edge_types():
+            warnings.warn(
+                "Edge types are used as Neo4j relationship types, "
+                "provided edge label will be ignored",
+                BlueGraphWarning)
+
     driver = generate_neo4j_driver(uri, username, password, driver)
 
     if pgframe is None:
@@ -86,6 +135,7 @@ def pgframe_to_neo4j(pgframe=None, uri=None, username=None, password=None,
             driver, node_label, edge_label, directed=directed)
 
     # Create nodes
+
     # Split nodes into batches
     batches = np.array_split(
         pgframe._nodes.index, math.ceil(pgframe.number_of_nodes() / batch_size))
@@ -95,51 +145,83 @@ def pgframe_to_neo4j(pgframe=None, uri=None, username=None, password=None,
         node_repr = []
         for index, properties in node_batch.to_dict("index").items():
             node_id = safe_node_id(index)
-            node_dict = ["id: '{}'".format(node_id)]
+            node_dict = [
+                "id: '{}'".format(node_id)
+            ]
             node_dict += _generate_property_repr(
                 properties, pgframe._node_prop_types)
             node_repr.append("{" + ", ".join(node_dict) + "}")
 
+        node_label_repr = f":{node_label}" if node_label else ""
+
         query = (
         f"""
         WITH [{", ".join(node_repr)}] AS batch
         UNWIND batch as individual
-        CREATE (n:{node_label})
+        CREATE (n{node_label_repr})
         SET n += individual
         """)
         execute(driver, query)
 
+    # Add node types to the Neo4j node labels
+    if node_types_as_labels:
+        with driver.session() as session:
+            for index, properties in pgframe._nodes.to_dict("index").items():
+                labels = labels_from_types(properties)
+                if len(labels) > 0:
+                    result = session.run(
+                        "MATCH (n {{id: '{}'}})\n".format(index) +
+                        "SET n:{}".format(":".join(labels))
+                    )
+
     # Create edges
-    # Split edges into batches
-    batches = np.array_split(
-        pgframe._edges.index, math.ceil(pgframe.number_of_edges() / batch_size))
-    for batch in batches:
-        edge_batch = pgframe._edges.loc[batch]
-        edge_repr = []
-        for (s, t), properties in edge_batch.to_dict("index").items():
-            edge_dict = [
-                "source: '{}'".format(safe_node_id(s)),
-                "target: '{}'".format(safe_node_id(t))
-            ]
-            edge_props = []
-            for k, v in properties.items():
-                quote = "'"
-                if pgframe._edge_prop_types[k] == "numeric":
-                    quote = ""
-                edge_props.append(f"{k}: {quote}{preprocess_value(v)}{quote}")
-            edge_dict.append("props: {{{}}}".format(
-                ', '.join(_generate_property_repr(
-                    properties, pgframe._edge_prop_types))))
-            edge_repr.append("{" + ", ".join(edge_dict) + "}")
-        query = (
-        f"""
-        WITH [{", ".join(edge_repr)}] AS batch
-        UNWIND batch as individual
-        MATCH (n:{node_label} {{id: individual["source"]}}), (m:{node_label} {{id: individual["target"]}})
-        CREATE (n)-[r:{edge_label}]->(m)
-        SET r += individual["props"]
-        """)
-        execute(driver, query)
+    custom_rel_types = edge_types_as_labels and pgframe.has_edge_types()
+    if custom_rel_types:
+        edge_labels = pgframe.edge_types(flatten=True)
+    else:
+        edge_labels = [edge_label]
+
+    for edge_label in edge_labels:
+        # Select edges of a given type, if applicable
+        edges = pgframe.edges(
+            raw_frame=True,
+            typed_by=edge_label if custom_rel_types else None)
+
+        # Split edges into batches
+        batches = np.array_split(
+            edges.index, math.ceil(edges.index.shape[0] / batch_size))
+        for batch in batches:
+            edge_batch = edges.loc[batch]
+            edge_repr = []
+            for (s, t), properties in edge_batch.to_dict("index").items():
+                edge_dict = [
+                    "source: '{}'".format(safe_node_id(s)),
+                    "target: '{}'".format(safe_node_id(t))
+                ]
+                edge_props = []
+                for k, v in properties.items():
+                    if k != "@type":
+                        quote = "'"
+                        if pgframe._edge_prop_types[k] == "numeric":
+                            quote = ""
+                        edge_props.append(
+                            f"{k}: {quote}{preprocess_value(v)}{quote}")
+                edge_dict.append("props: {{{}}}".format(
+                    ', '.join(_generate_property_repr(
+                        properties, pgframe._edge_prop_types))))
+                edge_repr.append("{" + ", ".join(edge_dict) + "}")
+
+            query = (
+            f"""
+            WITH [{", ".join(edge_repr)}] AS batch
+            UNWIND batch as individual
+            MATCH (n {{id: individual["source"]}})
+            WITH individual, n
+            OPTIONAL MATCH (m {{id: individual["target"]}})
+            CREATE (n)-[r:{edge_label}]->(m)
+            SET r += individual["props"]
+            """)
+            execute(driver, query)
 
     return Neo4jGraphView(driver, node_label, edge_label, directed=directed)
 
@@ -434,6 +516,8 @@ class Neo4jGraphProcessor(GraphProcessor):
         """Get a node/edge induced subgraph."""
         return Neo4jGraphView(
             self.driver, self.node_label, self.edge_label,
+            nodes_to_include=nodes_to_include,
+            edges_to_include=edges_to_include,
             nodes_to_exclude=nodes_to_exclude,
             edges_to_exclude=edges_to_exclude,
             directed=self.directed)
@@ -454,11 +538,16 @@ class Neo4jGraphView(object):
     TODO: make methods public
     """
     def __init__(self, driver, node_label,
-                 edge_label, nodes_to_exclude=None,
-                 edges_to_exclude=None, directed=True):
+                 edge_label,
+                 nodes_to_include=None, edges_to_include=None,
+                 nodes_to_exclude=None, edges_to_exclude=None,
+                 directed=True):
+        """Initialize an instance of Neo4jGraphView."""
         self.driver = driver
         self.node_label = node_label
         self.edge_label = edge_label
+        self.nodes_to_include = nodes_to_include
+        self.edges_to_include = edges_to_include
         self.nodes_to_exclude = nodes_to_exclude if nodes_to_exclude else []
         self.edges_to_exclude = edges_to_exclude if edges_to_exclude else []
         self.directed = directed
